#include"hash_fn.h"
#include"misc.h"
#include"bnb.h"

#include<limits.h>
#include<math.h>

extern int hashes;

int mhash(key_t key,size_t size) {
          return key % size;
    }

int rhash(key_t key,size_t size) {
		int bits = IBITS,rem = 0, pos = 0;

		while(size > 0) {
			size = (size >> 1) + (size & 1);
			rem = key % bits;
			if(key & (key_t)(1 << rem)) pos += size;
			rem++;
			key >>= rem;
			bits -= rem;
		}
		return pos;
	}

int  bhash(key_t key,size_t size) {
		int pos = 0,interval = size;
		unsigned bit = 1 << (IBITS - 1);

		while(interval > 0) {
			if(interval > 1)
			   interval = (interval & 1) + (interval >> 1);
			else 
			   interval = 0;
			if(key & bit) pos += interval;
			bit >>= 1;
		}
		return pos;
	 }

#ifndef LOG2
#define LOG2 (.6831471)
#endif

int phash(key_t key,size_t size) {
	return (int)((size + 1) * (log((long double)key) / (IBITS * LOG2)));
} 

#define KEYMUL (16)
int dhash(key_t key,size_t size) {
		int div = (int)((((double)KEYMUL) / size) * UINT_MAX);
		int pos;

		pos = (key / div) * KEYMUL +  (int)((((double)key)/UINT_MAX) * KEYMUL);
		return pos >= size ? size - 1 : pos;
	}

int rehash0(key_t key,size_t size) {
	   static int initial;
	   int pos;

	   if(hashes == 0) {
		 return pos = initial = key % size;
	   }
	   pos = ((key + (rehashstep * hashes)) % size);
	   return (pos != initial) ? pos : (-1);
	}

int rehash1(key_t key,size_t size) {
	   static int initial,pos;

	   if(hashes == 0) {
		 return pos = initial = key % size;
	   }
	   pos = ((key + (rehashstep * pos)) % size);
	   return (pos != initial) ? pos : (-1);
	}

int rehash2(key_t key,size_t size) {
	   static int pos,initial,step;

	   if(hashes == 0) {
		 step = 0;
		 return pos = initial = key % size;
	   }
	   step += pos;
	   pos = ((key + step) % size);
	   return (pos != initial) ? pos : (-1);
	}

int rehash3(key_t key,size_t size) {
	   static int pos,prev,pprev,step;

	   if(hashes == 0) {
		 step = 0;
		 pprev = prev = -1;
		 return pos = key % size;
	   }
	   step += hashes;
	   pprev = prev;
	   prev = pos;
	   pos = ((key + step) % size);
	   return (pos != pprev) ? pos : (-1);
	}

int rehash4(key_t key,size_t size) {
	   static int initial,pos;

	   if(hashes == 0) {
		 return pos = initial = key % size;
	   }
	   pos = ((key * (pos + 1)) % size);
	   return (pos != initial) ? pos : (-1);
	}

int rehash5(key_t key,size_t size) {
	   static int pos,initial;

	   if(hashes == 0) {
		 return pos = initial = key % size;
	   }
	   pos = ((key + pos) % size);
	   return (pos != initial) ? pos : (-1);
	}

int rehash6(key_t key,size_t size) {
	   static int initial;
	   int pos;

	   if(hashes == 0) {
		 return pos = initial = key % size;
	   }
	   pos = ((key * (hashes + 1)) % size);
	   return (pos != initial) ? pos : (-1);
	}

int prehashstep0(int size) {
       return isqrt(size) + 1;
    }

int prehash0(key_t key,size_t size) {
	   static int initial;
	   int pos;

	   if(hashes == 0) {
		 return pos = initial = key % size;
	   }
	   pos = ((pos + 1) * prehashstep) % size;
	   return (pos != initial) ? pos : (-1);
	}
